# This is a python implementation of the ForceAtlas2 plugin from Gephi
# intended to be used with networkx, but is in theory independent of
# it since it only relies on the adjacency matrix.  This
# implementation is based directly on the Gephi plugin:
#
# https://github.com/gephi/gephi/blob/master/modules/LayoutPlugin/src/main/java/org/gephi/layout/plugin/forceAtlas2/ForceAtlas2.java
#
# This is in contrast to an implementation I found in R (which uses
# strange parametrization) and to one in Python (which is based on the
# spring_layout).  The python implementation did not seem to work, and
# I did not try the R implementation because I didn't want to spend a
# long time setting up a messy pipe system only to find that the
# algorithm was as wrong as the python one.
#
# For simplicity and for keeping code in sync with upstream, I have
# reused as many of the variable/function names as possible, even when
# they are in a more java-like style (e.g. camalcase)
#
# I wrote this because I was unable to find a graph layout algorithm
# in Python that clearly showed modular structure.
#
# NOTES: Currently, this only works for unweighted, undirected graphs.
#
# Copyright 2016 Max Shinn <mws41@cam.ac.uk>
#
# Available under the GPLv3

"""
Author: Max Shinn 2016
Modified on 20.4.17, Lenz Baumann
E-Mail: lnzbmnn@gmail.com
"""

import random
from math import sqrt

import numpy

from twitter_data_equipment import fa2util


def forceatlas2(
        G,  # a graph, in 2D numpy ndarray format
        pos=None,  # Vector of initial positions
        niter=100,  # Number of times to iterate the main loop

        # Behavior alternatives
        outboundAttractionDistribution=False,  # "Dissuade hubs" # NOT (fully) IMPLEMENTED
        linlogMode=False,
        adjustSizes=False,
        edgeWeightInfluence=1,

        # Performance
        jitterTolerance=1.0,  # "Tolerance"

        # Tuning
        scalingRatio=2.0,
        strongGravityMode=False,
        gravity=1.0,
        sizes=None
):
    # Check our assumptions
    assert isinstance(G, numpy.ndarray), "G is not a numpy ndarray"
    assert G.shape == (G.shape[0], G.shape[0]), "G is not 2D square"
    assert numpy.all(G.T == G), "G is not symmetric.  Currently only undirected graphs are supported"
    assert isinstance(pos, numpy.ndarray) or (pos is None), "Invalid node positions"
    assert len(G) > 0, "Graph is empty"
    # assert outboundAttractionDistribution == linLogMode == adjustSizes == barnesHutOptimize == False, "You selected a feature that has not been implemented yet..."



    # Initializing, ala initAlgo()
    # ============================

    # speed and speedEfficiency describe a scaling factor of dx and dy
    # before x and y are adjusted.  These are modified as the
    # algorithm runs to help ensure convergence.
    speed = 1
    speedEfficiency = 1

    # Put nodes into a data structure we can understand
    nodes = []

    for i in range(0, G.shape[0]):
        n = fa2util.Node()
        n.mass = 1 + numpy.sum(G[i])
        n.old_dx = 0
        n.old_dy = 0
        n.dx = 0
        n.dy = 0
        if sizes is not None:
            n.size = int(sizes[i])
        if pos is None:
            n.x = random.random()
            n.y = random.random()
        else:
            n.x = pos[i][0]
            n.y = pos[i][1]
        nodes.append(n)

    # Put edges into a data structure we can understand
    edges = []
    es = numpy.asarray(G.nonzero()).T
    for e in es:  # Iterate through edges
        if e[1] <= e[0]: continue  # Avoid duplicate edges
        edge = fa2util.Edge()
        edge.node1 = e[0]  # The index of the first node in `nodes`
        edge.node2 = e[1]  # The index of the second node in `nodes`
        edge.weight = G[tuple(e)]
        edges.append(edge)

    # Main loop, i.e. goAlgo()
    # ========================

    # Each iteration of this loop reseprensts a call to goAlgo().

    helper_ind = 0
    for _i in range(0, niter):
        #helper_ind+=1
        #print(str(helper_ind/niter*100))
        for n in nodes:
            n.old_dx = n.dx
            n.old_dy = n.dy
            n.dx = 0
            n.dy = 0

        # Barnes Hut optimization step goes here...

        if outboundAttractionDistribution:
            outboundAttCompensation = numpy.mean([n.mass for n in nodes])

        fa2util.apply_repulsion(nodes, scalingRatio, adjustSizes=adjustSizes)
        fa2util.apply_gravity(nodes, gravity, scalingRatio, useStrongGravity=strongGravityMode, adjustSizes=adjustSizes)
        fa2util.apply_attraction(nodes, edges, scalingRatio, edgeWeightInfluence, adjustSizes=adjustSizes, linlogMode=linlogMode)

        # Auto adjust speed.
        totalSwinging = 0.0  # How much irregular movement
        totalEffectiveTraction = 0.0  # How much useful movement
        for n in nodes:
            swinging = sqrt((n.old_dx - n.dx) * (n.old_dx - n.dx) + (n.old_dy - n.dy) * (n.old_dy - n.dy))
            totalSwinging += n.mass * swinging
            totalEffectiveTraction += .5 * n.mass * sqrt(
                (n.old_dx + n.dx) * (n.old_dx + n.dx) + (n.old_dy + n.dy) * (n.old_dy + n.dy))

        # Optimize jitter tolerance.  The 'right' jitter tolerance for
        # this network. Bigger networks need more tolerance. Denser
        # networks need less tolerance. Totally empiric.
        estimatedOptimalJitterTolerance = .05 * sqrt(len(nodes))
        minJT = sqrt(estimatedOptimalJitterTolerance)
        maxJT = 10
        jt = jitterTolerance * max(minJT, min(maxJT, estimatedOptimalJitterTolerance * totalEffectiveTraction / (
            len(nodes) * len(nodes))))

        minSpeedEfficiency = 0.05

        if totalSwinging / totalEffectiveTraction > 2.0:
            if speedEfficiency > minSpeedEfficiency:
                speedEfficiency *= .5
            jt = max(jt, jitterTolerance)

        targetSpeed = jt * speedEfficiency * totalEffectiveTraction / totalSwinging

        if totalSwinging > jt * totalEffectiveTraction:
            if speedEfficiency > minSpeedEfficiency:
                speedEfficiency *= .7
        elif speed < 1000:
            speedEfficiency *= 1.3

        maxRise = .5
        speed = speed + min(targetSpeed - speed, maxRise * speed)

        if adjustSizes:
            for n in nodes:
                swinging = n.mass * sqrt((n.old_dx - n.dx) * (n.old_dx - n.dx) + (n.old_dy - n.dy) * (n.old_dy - n.dy))
                factor = 0.1 * speed / (1.0 + sqrt(speed * swinging))
                df = sqrt(pow(n.dx, 2) + pow(n.dy, 2))
                factor = min(factor * df, 10.0) / df
                n.x += n.dx * factor
                n.y += n.dy * factor
        else:
            for n in nodes:
                swinging = n.mass * sqrt((n.old_dx - n.dx) * (n.old_dx - n.dx) + (n.old_dy - n.dy) * (n.old_dy - n.dy))
                factor = speed / (1.0 + sqrt(speed * swinging))
                n.x += n.dx * factor
                n.y += n.dy * factor
    return [(n.x, n.y) for n in nodes]


def forceatlas2_networkx_layout(G, pos=None, sizes=None, **kwargs):
    import networkx
    assert isinstance(G, networkx.classes.graph.Graph), "Not a networkx graph"
    assert isinstance(pos, dict) or (pos is None), "pos must be specified as a dictionary, as in networkx"
    M = numpy.asarray(networkx.to_numpy_matrix(G))
    if pos is None:
        l = forceatlas2(M, pos=None, sizes=sizes, **kwargs)
    else:
        poslist = numpy.asarray([pos[i] for i in G.nodes()])
        l = forceatlas2(M, pos=poslist, **kwargs)
    return dict(zip(G.nodes(), l))



"""
This module helps hydrating a file containing a list of tweet-ids. It gets the file with the tweet-ids as input
and produces a Json-File as output.
"""

from twarc import Twarc
import json




def hydrate_file(input_path, output_path, key_dict):
    """
    This function gets a file with tweet-ids, an ouput-path for the Json-File and a dict containing the 
    twitter-credentials as input, and produces a Json-File containing the hydrated tweets.
    The method is based on the library 'Twarc' and the call to the method 'Twarc.hydrate'
    
    :param str input_path: The path leading to the input-file. The input-file must have a header and then contain a tweet-id one at a line.
    
    :param str output_path: The path leading to the output-file. The output-file gets generated if not existent and overwriten if existent.
    
    :param dict of (str,str) key_dict: The key_dict must contain for keys containing the twitter login information and be of the following form:
    {
    'consumer_key':'<consumer_key>',
    'consumer_secret':'<consumer_secret>',
    'access_token':'<access_token>',
    'access_token_secret':'<access_token_secret>'
    }
    
    :return: 
    
    """
    i = open(input_path)
    t = Twarc(key_dict['consumer_key'], key_dict['consumer_secret'], key_dict['access_token'],
              key_dict['access_token_secret'])
    with open(output_path, 'w') as outfile:
        for tweet in t.hydrate(i):
            outfile.write(json.dumps(tweet) + '\n')




"""
Module for calculating various percentages as well as the top X entries of 
hashtags, users, urls, media and locations that appear in a dataframe
"""

from collections import Counter


def get_top_n_hashtags(n_positions, dataframe, give_freq=False):
    """
    Returns the most used hashtags in the given dataframe, amount is specified in the n_n_positions parameter,
    and the give_freq parameter specifies whether the frequency of each hashtag should be returned.
    
    :param int n_positions: Number of most used hashtags should be returned. E.g. Top 10 or Top 5
    
    :param pd.Dataframe dataframe: Dataframe containing twitter data
    
    :param boolean give_freq: Whether to return the frequency or not
    
    :return list: Containing the specified amount of hashtags either including their exact frequency or not 
    
    """
    df = dataframe[dataframe.hashtags != '[]']
    hashtag_list = list()
    for row in df.itertuples():
        try:
            hashtags = eval(row.hashtags)
            for hashtag in hashtags:
                hashtag_list.append(hashtag)
        except:
            pass
    hashtags_to_freq = dict(Counter(hashtag_list))
    sorted_hash_freq = sorted(hashtags_to_freq.items(), key=lambda x: x[1], reverse=True)
    if give_freq:
        return sorted_hash_freq[0:n_positions]
    else:
        return [x[0] for x in sorted_hash_freq[0:n_positions]]


def get_all_hashtags(dataframe):
    """
    Returns all hashtags in the given dataframe without their frequencies

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return list: Containing the all hashtags

    """
    df = dataframe[dataframe.hashtags != '[]']
    hashtag_list = list()
    for row in df.itertuples():
        try:
            hashtags = eval(row.hashtags)
            for hashtag in hashtags:
                hashtag_list.append(hashtag)
        except:
            pass

    hashtags_to_freq = dict(Counter(hashtag_list))
    sorted_hash_freq = sorted(hashtags_to_freq.items(), key=lambda x: x[1], reverse=True)
    return [x[0] for x in sorted_hash_freq]


def get_top_n_users(n_positions, dataframe, give_freq=False):
    """
    Returns the most active users (measured on how many tweets they tweeted) in the given dataframe, 
    amount is specified in the n_n_positions parameter and the give_freq parameter specifies whether the frequency of 
    each hashtag should be returned.

    :param int n_positions: Number of most active user should be returned. E.g. Top 10 or Top 5

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :param boolean give_freq: Whether to return the frequency or not

    :return list: Containing the specified amount of users either including their exact frequency or not 

    """
    users_list = list()
    for row in dataframe.itertuples():
        user = row.u_id
        name = row.screen_name
        users_list.append((user,name))

    users_to_freq = dict(Counter(users_list))
    sorted_users_freq = sorted(users_to_freq.items(), key=lambda x: x[1], reverse=True)
    if give_freq:
        return sorted_users_freq[0:n_positions]
    else:
        return [x[0] for x in sorted_users_freq[0:n_positions]]


def get_top_n_urls(n_positions, dataframe, give_freq=False):
    """
    Returns the most tweeted urls in the given dataframe, amount is specified in the n_n_positions parameter 
    and the give_freq parameter specifies whether the frequency of each url should be returned.

    :param int n_positions: Number of most tweeted url should be returned. E.g. Top 10 or Top 5

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :param boolean give_freq: Whether to return the frequency or not

    :return list: Containing the specified amount of urls either including their exact frequency or not 

    """
    df = dataframe[dataframe.urls != '[]']
    urls_list = list()
    for row in df.itertuples():
        try:
            urls = eval(row.urls)
            for url in urls:
                urls_list.append(url)
        except:
            pass

    urls_to_freq = dict(Counter(urls_list))
    sorted_urls_freq = sorted(urls_to_freq.items(), key=lambda x: x[1], reverse=True)
    if give_freq:
        return sorted_urls_freq[0:n_positions]
    else:
        return [x[0] for x in sorted_urls_freq[0:n_positions]]


def get_top_n_media(n_positions, dataframe, give_freq=False):
    """
    Returns the most tweeted media-urls in the given dataframe, 
    amount is specified in the n_n_positions parameter and the give_freq parameter specifies whether the frequency of 
    each media-url should be returned.

    :param int n_positions: Number of most tweeted media-urls should be returned. E.g. Top 10 or Top 5

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :param boolean give_freq: Whether to return the frequency or not

    :return list: Containing the specified amount of media-urls either including their exact frequency or not 

    """
    df = dataframe[dataframe.media == 1]
    media_url_list = list()
    for row in df.itertuples():
        try:
            medias = eval(row.mediaurls)
            for media in medias:
                media_url_list.append(media)
        except:
            pass

    media_to_freq = dict(Counter(media_url_list))
    sorted_mediaurls_freq = sorted(media_to_freq.items(), key=lambda x: x[1], reverse=True)
    if give_freq:
        return sorted_mediaurls_freq[0:n_positions]
    else:
        return [x[0] for x in sorted_mediaurls_freq[0:n_positions]]


def get_top_n_locations(n_positions,dataframe,unit_of_location='States',give_freq = False):
    """
    Returns the most geo-locations that tweets where sent from in the given dataframe, 
    amount is specified in the n_n_positions parameter and the give_freq parameter specifies whether the frequency of 
    each location should be returned.

    :param int n_positions: Number of most geo-locations that tweets where sent from should be returned. E.g. Top 10 or Top 5

    :param pd.Dataframe dataframe: Dataframe containing twitter data
    
    :param str unit_of_location: The basic geographic unit that is taken to identify wheter to tweets were tweeted from the same place or not

    :param boolean give_freq: Whether to return the frequency or not

    :return list: Containing the specified amount of Locations either including their exact frequency or not 

    """
    if 'has_stats' in dataframe:
        df = dataframe[dataframe.has_stats == 1]
        state_fips_list = list()
        if len(df) > 0:
            for row in df.itertuples():
                try:
                    if unit_of_location == 'States':
                        fips = int(row.State_FIPS)
                        name = str(row.State_Name)
                    elif unit_of_location == 'Counties':
                        fips = int(row.County_FIPS)
                        name = str(row.County_Name)
                except Exception as e:
                    pass
                state_fips_list.append((fips, name))
            state_fips_to_freq = dict(Counter(state_fips_list))
            sorted_state_fips_freq = sorted(state_fips_to_freq.items(), key=lambda x: x[1], reverse=True)
            if give_freq:
                return sorted_state_fips_freq[0:n_positions]
            else:
                return [x[0][1] for x in sorted_state_fips_freq[0:n_positions]]
        else:
            return []
    else:
        return []


def get_perc_of_geo_info(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that contain geo-info
    
    :param pd.Dataframe dataframe: Dataframe containing twitter data
    
    :return: Either string telling the error or integer telling the percentage
    
    """
    if 'is_geo' in dataframe:
        df_geo = dataframe[dataframe.is_geo == 1]
        if len(df_geo) > 0:
            return round(len(df_geo) / len(dataframe) * 100,1)
        else:
            return 'No geographical Information in dataset'
    else:
        return 'No geographical Information in dataset'


def get_perc_of_stats_info(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that contain demographical information

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return: Either string telling the error or integer telling the percentage

    """
    if 'has_stats' in dataframe:
        df_stats = dataframe[dataframe.has_stats == 1]
        if len(df_stats) > 0:
            return round(len(df_stats) / len(dataframe) * 100,1)
        else:
            return 'No demographic statistic in dataset'
    else:
        return 'No demographic statistic in dataset'


def get_perc_of_media(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that contain media-urls

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return: Either string telling the error or integer telling the percentage

    """
    if 'media' in dataframe:
        df_stats = dataframe[dataframe.media == 1]
        if len(df_stats) > 0:
            return round(len(df_stats) / len(dataframe) * 100,1)
        else:
            return 'No media in dataframe'
    else:
        return 'No media in dataframe'


def get_perc_of_qt(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that are quotes

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return: Either string telling the error or integer telling the percentage

    """
    if 'QT' in dataframe:
        df_stats = dataframe[dataframe.QT == 1]
        if len(df_stats) > 0:
            return round(len(df_stats) / len(dataframe) * 100,1)
        else:
            return 'No quoting in dataset'
    else:
        return 'No quoting in dataset'


def get_perc_of_rt(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that are retweets

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return: Either string telling the error or integer telling the percentage

    """
    if 'RT' in dataframe:
        df_stats = dataframe[dataframe.RT == 1]
        if len(df_stats) > 0:
            return round(len(df_stats) / len(dataframe) * 100,1)
        else:
            return 'No retweets in dataset'
    else:
        return 'No retweets in dataset'


def get_perc_of_replies(dataframe):
    """
    Calculates the percentage of all tweets in the given dataframe that are replies

    :param pd.Dataframe dataframe: Dataframe containing twitter data

    :return: Either string telling the error or integer telling the percentage

    """
    if 'is_repy' in dataframe:
        df_stats = dataframe[dataframe.is_reply == 1]
        if len(df_stats > 0):
            return round(len(df_stats) / len(dataframe) * 100,1)
        else:
            return 'No replies in dataset'
    else:
        return 'No replies in dataset'


def analyze_dataframe(dataframe, give_freq):
    """
    Analyzes the dataframe through application of all specified methods above.
    
    :param pandas.Dataframe dataframe: Dataframe containing twitter data
    
    :param bool give_freq: Whether frequency should be returned or not
    
    :return dict: Containing all information on hashtags, users, media, urls, geo-info and demographic-info
    
    """
    analyze_dict = dict()
    analyze_dict['hashtags'] = get_top_n_hashtags(10, dataframe, give_freq=give_freq)
    analyze_dict['urls'] = get_top_n_urls(10, dataframe, give_freq=give_freq)
    analyze_dict['media'] = get_top_n_media(10, dataframe, give_freq=give_freq)
    analyze_dict['users'] = get_top_n_users(10, dataframe, give_freq=give_freq)
    analyze_dict['state_fips'] = get_top_n_locations(10,dataframe,give_freq=give_freq)
    analyze_dict['geo_perc'] = get_perc_of_geo_info(dataframe)
    analyze_dict['stats_perc'] = get_perc_of_stats_info(dataframe)
    analyze_dict['media_perc'] = get_perc_of_media(dataframe)
    analyze_dict['qt_perc'] = get_perc_of_qt(dataframe)
    analyze_dict['rt_perc'] = get_perc_of_rt(dataframe)
    analyze_dict['reply_perc'] = get_perc_of_replies(dataframe)
    return analyze_dict



# This file allows separating the most CPU intensive routines from the
# main code.  This allows them to be optimized with Cython.  If you
# don't have Cython, this will run normally.  However, if you use
# Cython, you'll get speed boosts from 10-100x automatically.
#
# The only catch is that IF YOU MODIFY THIS FILE, YOU MUST ALSO MODIFY
# fa2util.pyd TO REFLECT ANY CHANGES IN FUNCTION DEFINITIONS!
#
# Copyright 2016 Max Shinn <mws41@cam.ac.uk>
#
# Available under the GPLv3

"""
Author: Max Shinn 2016
Modified on 20.4.17, Lenz Baumann
E-Mail: lnzbmnn@gmail.com
"""

from math import sqrt
from math import log


# This will substitute for the nLayout object
class Node:
    def __init__(self):
        self.mass = 0
        self.old_dx = 0
        self.old_dy = 0
        self.dx = 0
        self.dy = 0
        self.x = 0
        self.y = 0
        self.size = 0


# This is not in the original java code, but it makes it easier to
# deal with edges.
class Edge:
    def __init__(self):
        self.node1 = -1
        self.node2 = -1
        self.weight = 0


class Region:
    def __init__(self, nodes):
        self.mass = 0
        self.mass_center_x = 0
        self.mass_center_y = 0
        self.nodes = (Node)
        nodes
        self.subregions = []
        self.size = 0
        self.update_mass_and_geometry()

    def update_mass_and_geometry(self):
        if len(self.nodes) > 1:
            mass = 0
            mass_sum_y = 0
            mass_sum_x = 0
            for node in self.nodes:
                mass += node.mass
                mass_sum_x += node.x * node.mass
                mass_sum_y += node.y * node.mass

            mass_center_x = mass_sum_x / mass
            mass_center_y = mass_sum_y / mass

            for node in self.nodes:
                distance = sqrt((node.x - mass_center_x) * (node.x - mass_center_x) +
                                (node.y - mass_center_y) * (node.y - mass_center_y))
                size = max(size, 2 * distance)

    def build_subregions(self):
        if len(self.nodes) > 1:
            left_nodes = []
            right_nodes = []

            for node in self.nodes:
                if node.x < self.mass_center_x:
                    left_nodes.append(node)
                else:
                    right_nodes.append(node)

            top_left_nodes = []
            bottom_left_nodes = []
            for node in left_nodes:
                if node.y < self.mass_center_y:
                    top_left_nodes.append(node)
                else:
                    bottom_left_nodes.append(node)

            top_right_nodes = []
            bottom_right_nodes = []
            for node in right_nodes:
                if node.y < self.mass_center_y:
                    top_right_nodes.append(node)
                else:
                    bottom_right_nodes.append(node)


            if len(top_left_nodes) > 0:
                if len(top_right_nodes) < len(self.nodes):
                    self.subregions.append(Region(top_left_nodes))
                else:
                    for node in top_left_nodes:
                        one_node_list = []
                        one_node_list.append(node)
                        self.subregions.append(Region(one_node_list))
            if len(bottom_left_nodes) > 0:
                if len(bottom_left_nodes) < len(self.nodes):
                    self.subregions.append(Region(bottom_left_nodes))
                else:
                    for node in bottom_left_nodes:
                        one_node_list = []
                        one_node_list.append(node)
                        self.subregions.append(Region(one_node_list))
            if len(top_right_nodes) > 0:
                if len(top_right_nodes) < len(self.nodes):
                    self.subregions.append(Region(top_right_nodes))
                else:
                    for node in top_right_nodes:
                        one_node_list = []
                        one_node_list.append(node)
                        self.subregions.append(Region(one_node_list))
            if len(bottom_right_nodes) > 0:
                if len(bottom_right_nodes) < len(self.nodes):
                    self.subregions.append(Region(bottom_right_nodes))
                else:
                    for node in bottom_right_nodes:
                        one_node_list = []
                        one_node_list.append(node)
                        self.subregions.append(Region(one_node_list))

            for region in self.subregions:
                region.build_subregions()





# Here are some functions from ForceFactor.java
# =============================================

# Repulsion function.  `n1` and `n2` should be nodes.  This will
# adjust the dx and dy values of `n1` (and optionally `n2`).  It does
# not return anything.
def linRepulsion(n1, n2, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    distance = xDist * xDist + yDist * yDist  # Distance squared

    if distance > 0:
        factor = coefficient * n1.mass * n2.mass / distance
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor


def linRepulsion_anticollision(n1, n2, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    # distance = sqrt(xDist * xDist + yDist * yDist)
    # distance2 = distance - n1.size - n2.size
    distance = sqrt(xDist * xDist + yDist * yDist) - n1.size - n2.size


    if distance > 0:
        factor = coefficient * n1.mass * n2.mass / (distance * distance)
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor
    elif distance < 0:
        factor = 1 * coefficient * n1.mass * n2.mass
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor


# Gravity repulsion function.  For some reason, gravity was included
# within the linRepulsion function in the original gephi java code,
# which doesn't make any sense (considering a. gravity is unrelated to
# nodes repelling each other, and b. gravity is actually an
# attraction).
def linGravity(n, g, coefficient):
    xDist = n.x
    yDist = n.y
    distance = sqrt(xDist * xDist + yDist * yDist)

    if distance > 0:
        factor = coefficient * n.mass * g / distance
        n.dx -= xDist * factor
        n.dy -= yDist * factor


def linGravity_anticollision(n, g, coefficient):
    xDist = n.x
    yDist = n.y
    distance = sqrt(xDist * xDist + yDist * yDist)

    if distance > 0:
        factor = coefficient * n.mass * g / distance
        n.dx -= xDist * factor
        n.dy -= yDist * factor


# Strong gravity force function.  `n` should be a node, and `g`
# should be a constant by which to apply the force.
def strongGravity(n, g, coefficient):
    xDist = n.x
    yDist = n.y

    if xDist != 0 and yDist != 0:
        factor = coefficient * n.mass * g
        n.dx -= xDist * factor
        n.dy -= yDist * factor


# Attraction function.  `n1` and `n2` should be nodes.  This will
# adjust the dx and dy values of `n1` (and optionally `n2`).  It does
# not return anything.
def linAttraction(n1, n2, e, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    factor = -coefficient * e
    n1.dx += xDist * factor
    n1.dy += yDist * factor
    n2.dx -= xDist * factor
    n2.dy -= yDist * factor


def linAttraction_anticollision(n1, n2, e, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    distance = sqrt(xDist * xDist + yDist * yDist) - n1.size - n2.size

    if (distance > 0):
        factor = -coefficient * e
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor

def logAttraction(n1, n2, e, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    distance = sqrt(xDist * xDist + yDist * yDist)

    if (distance > 0):
        factor = -coefficient * e * log(1 + distance) / distance
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor


def logAttraction_anticollision(n1, n2, e, coefficient):
    xDist = n1.x - n2.x
    yDist = n1.y - n2.y
    distance = sqrt(xDist * xDist + yDist * yDist) - n1.size - n2.size

    if (distance > 0):
        factor = -coefficient * e * log(1 + distance) / distance
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor


# The following functions iterate through the nodes or edges and apply
# the forces directly to the node objects.  These iterations are here
# instead of the main file because Python is slow with loops.  Where
# relevant, they also contain the logic to select which version of the
# force function to use.
def apply_repulsion(nodes, coefficient, adjustSizes):
    if adjustSizes:
        for i in range(0, len(nodes)):
            for j in range(0, i):
                linRepulsion_anticollision(nodes[i], nodes[j], coefficient)
    else:
        for i in range(0, len(nodes)):
            for j in range(0, i):
                linRepulsion(nodes[i], nodes[j], coefficient)


def apply_gravity(nodes, gravity, scalingRatio, useStrongGravity, adjustSizes):
    if adjustSizes:
        if not useStrongGravity:
            for i in range(0, len(nodes)):
                linGravity_anticollision(nodes[i], gravity / scalingRatio, scalingRatio)
        else:
            for i in range(0, len(nodes)):
                strongGravity(nodes[i], gravity / scalingRatio, scalingRatio)
    else:
        if not useStrongGravity:
            for i in range(0, len(nodes)):
                linGravity(nodes[i], gravity / scalingRatio, scalingRatio)
        else:
            for i in range(0, len(nodes)):
                strongGravity(nodes[i], gravity / scalingRatio, scalingRatio)


def apply_attraction(nodes, edges, coefficient, edgeWeightInfluence, adjustSizes, linlogMode):
    if linlogMode:
        if adjustSizes:
            if edgeWeightInfluence == 0:
                for edge in edges:
                    logAttraction_anticollision(nodes[edge.node1], nodes[edge.node2], 1, coefficient)
            elif edgeWeightInfluence == 1:
                for edge in edges:
                    logAttraction_anticollision(nodes[edge.node1], nodes[edge.node2], edge.weight, coefficient)
            else:
                for edge in edges:
                    logAttraction_anticollision(nodes[edge.node1], nodes[edge.node2],
                                                pow(edge.weight, edgeWeightInfluence), coefficient)
        else:
            if edgeWeightInfluence == 0:
                for edge in edges:
                    logAttraction(nodes[edge.node1], nodes[edge.node2], 1, coefficient)
            elif edgeWeightInfluence == 1:
                for edge in edges:
                    logAttraction(nodes[edge.node1], nodes[edge.node2], edge.weight, coefficient)
            else:
                for edge in edges:
                    logAttraction(nodes[edge.node1], nodes[edge.node2], pow(edge.weight, edgeWeightInfluence),
                                  coefficient)
    else:
        if adjustSizes:
            if edgeWeightInfluence == 0:
                for edge in edges:
                    linAttraction_anticollision(nodes[edge.node1], nodes[edge.node2], 1, coefficient)
            elif edgeWeightInfluence == 1:
                for edge in edges:
                    linAttraction_anticollision(nodes[edge.node1], nodes[edge.node2], edge.weight, coefficient)
            else:
                for edge in edges:
                    linAttraction_anticollision(nodes[edge.node1], nodes[edge.node2],
                                                pow(edge.weight, edgeWeightInfluence), coefficient)
        else:
            if edgeWeightInfluence == 0:
                for edge in edges:
                    linAttraction(nodes[edge.node1], nodes[edge.node2], 1, coefficient)
            elif edgeWeightInfluence == 1:
                for edge in edges:
                    linAttraction(nodes[edge.node1], nodes[edge.node2], edge.weight, coefficient)
            else:
                for edge in edges:
                    linAttraction(nodes[edge.node1], nodes[edge.node2], pow(edge.weight, edgeWeightInfluence),
                                  coefficient)


try:
    import cython

    if not cython.compiled:
        print("Warning: uncompiled fa2util module.  Compile with cython for a 10-100x speed boost.")
except:
    print("No cython detected.  Install cython and compile the fa2util module for a 10-100x speed boost.")



from setuptools import setup

setup(name='twitter_data_equipment',
      version='1.4.8',
      description='A package that wraps tools for harvesting, hydrating, cleaning and visualize twitter data',
      url='https://github.com/HolzmanoLagrene/FacharbeitLenzBaumannHS16FS17',
      author='Lenz Baumann',
      author_email='lnzbmnn@gmail.com',
      license='MIT',
      packages=['twitter_data_equipment'],
      include_package_data=True,
      zip_safe=False)

